#!/bin/python3
#%%
import numpy as np
import pandas as pd
from tools.data_processing.basic_data_tools import multiprocessing_pool_map, save_and_reset_index

#%%
# re-write

def rewrited_stable_cumsum(arr, axis=None, rtol=1e-05, atol=1e-08):
    out = np.cumsum(arr, axis=axis, dtype=np.float64)
    return out


def rewrited_binary_clf_curve(y_true, y_score, pos_label=None, sample_weight=None):
    pos_label = 1
    if isinstance(y_true, pd.Series):
        y_true = y_true.to_numpy()
    if isinstance(y_score, pd.Series):
        y_score = y_score.to_numpy()
    y_true = y_true == pos_label
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]
    if sample_weight is not None:
        weight = sample_weight[desc_score_indices]
    else:
        weight = 1.0
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]
    tps = rewrited_stable_cumsum(y_true * weight)[threshold_idxs]
    if sample_weight is not None:
        fps = rewrited_stable_cumsum((1 - y_true) * weight)[threshold_idxs]
    else:
        fps = 1 + threshold_idxs - tps
    return fps, tps, y_score[threshold_idxs]


def rewrited_roc_curve(y_true, y_score, pos_label=None, sample_weight=None, drop_intermediate=True):
    fps, tps, thresholds = rewrited_binary_clf_curve(
        y_true, y_score, pos_label=pos_label, sample_weight=sample_weight
    )
    if drop_intermediate and len(fps) > 2:
        optimal_idxs = np.where(
            np.r_[True, np.logical_or(np.diff(fps, 2), np.diff(tps, 2)), True]
        )[0]
        fps = fps[optimal_idxs]
        tps = tps[optimal_idxs]
        thresholds = thresholds[optimal_idxs]
    tps = np.r_[0, tps]
    fps = np.r_[0, fps]
    thresholds = np.r_[thresholds[0] + 1, thresholds]
    fpr = fps / fps[-1]
    tpr = tps / tps[-1]
    return fpr, tpr, thresholds


def rewrited_auc(x, y):
    direction = 1
    dx = np.diff(x)
    if np.any(dx < 0):
        if np.all(dx <= 0):
            direction = -1
    area = direction * np.trapz(y, x)
    if isinstance(area, np.memmap):
        area = area.dtype.type(area)
    return area


def rewrited_binary_roc_auc_score(y_true, y_score, sample_weight=None):
    # if isinstance(y_true, pd.Series):
    #     y_true = y_true.to_numpy()
    # if isinstance(y_score, pd.Series):
    #     y_score = y_score.to_numpy()
    fpr, tpr, _ = rewrited_roc_curve(y_true, y_score, sample_weight=sample_weight)
    ret = rewrited_auc(fpr, tpr)
    ret = float(ret)
    return ret

roc_curve = rewrited_roc_curve
roc_auc_score = rewrited_binary_roc_auc_score

def ks_score(y, y_proba, sample_weight = None):
    fpr, tpr = roc_curve(y, y_proba, sample_weight=sample_weight)[:2]
    r = np.abs(tpr - fpr).max()
    return r


#%%

def series_auc_and_ks_score(
        y_series: pd.Series, 
        x_series: pd.Series,
        metric = 'auc_and_ks',
        dropna = False,
        fillna = 'mean',
        mode = 'union',
        to_frame = False,
        T = False
        ):
    
    if mode == 'union':
        data_df = pd.DataFrame({
            'y': y_series,
            'x': x_series
        })
        y_series = data_df['y']
        x_series = data_df['x']
    elif mode == 'intersection':
        idx = y_series.index.intersection(x_series.index)
        y_series = y_series[idx]
        x_series = x_series[idx]

    if dropna:
        idx = y_series.notna() & x_series.notna()
        y_series = y_series[idx]
        x_series = x_series[idx]
    else:
        idx = y_series.notna() 
        y_series = y_series[idx]
        x_series = x_series[idx]

    if fillna is not None:
        if isinstance(fillna, str) and fillna == 'mean':
            fillna = x_series.mean()
        x_series = x_series.astype(float).fillna(fillna)

    if metric == 'auc_and_ks':
        result = pd.Series(index=['auc', 'ks'])
        result.index.name = 'metric'
        result.name = 'stat'
    elif metric in ['auc', 'ks']:
        result = float('nan')
        
    if y_series.nunique() == 2 and x_series.notna().any():
        if metric == 'auc':
            result = roc_auc_score(y_series, x_series)
        elif metric == 'ks':
            result = ks_score(y_series, x_series)
        elif metric == 'auc_and_ks':
            result = pd.Series({
                'auc': roc_auc_score(y_series, x_series),
                'ks': ks_score(y_series, x_series),
            })

    if isinstance(result, pd.Series) and to_frame:
        result = result.to_frame()
        if T :
            result = result.T
            
    return result


import functools

series_auc_score = functools.partial(series_auc_and_ks_score, metric = 'auc')
series_ks_score = functools.partial(series_auc_and_ks_score, metric = 'ks')


def dataframe_auc_and_ks_score(
    data_df: pd.DataFrame, 
    y_col: str = 'y',
    x_cols: list = None,
    disable_multiprocessing = False
    ):
    y_series = data_df[y_col]
    if x_cols is None:
        x_df = data_df.select_dtypes(include='number').drop(columns=[y_col])
    else:
        x_df = data_df[x_cols]
    func = functools.partial(
        series_auc_and_ks_score, 
        y_series,
        to_frame = True,
        T = True
        )
    dfs = multiprocessing_pool_map(
        func=func, iterable=x_df,
        disable_multiprocessing=disable_multiprocessing
    )
    stat_df = pd.concat(dfs)
    stat_df = stat_df.droplevel(level=1)
    stat_df.index.name = 'feature_name'
    # stat_df.columns.name = 'metric'
    return stat_df

#%%

# def get_describe_by_month_df(xy_df: pd.DataFrame):
#     gp = xy_df.groupby(by=['month'])['y_proba']
#     df = gp.describe()
#     df.insert(0, 'size', gp.size())
#     df.loc['total'] = xy_df['y_proba'].describe()
#     df.loc['total', 'size'] = len(xy_df)
#     df.insert(2, 'coverage', df['count'] / df['size'])
#     df['size'] = df['size'].astype(int)
#     df['count'] = df['count'].astype(int)
#     df = save_and_reset_index(df)
#     return df

# def roc_auc_score_with_nan(y: pd.Series, y_score: pd.Series):
#     idx = (~y.isna()) & (~y_score.isna())
#     if idx.sum() > 1 and len(y[idx].unique()) > 1 > 0:
#         ret = roc_auc_score(y[idx], y_score[idx])
#     else:
#         ret = float('nan')
#     return ret

# def ks_score_with_nan(y: pd.Series, y_score: pd.Series):
#     idx = (~y.isna()) & (~y_score.isna())
#     if idx.sum() > 1 and len(y[idx].unique()) > 1 > 0:
#         ret = ks_score(y[idx], y_score[idx])
#     else:
#         ret = float('nan')
#     return ret


# def get_auc_and_ks_by_data_split(xy_df: pd.DataFrame, rename = None):
#     idx = (~xy_df['data_split'].isna())
#     data_df = xy_df.loc[idx, :].copy()
#     gp = data_df.groupby(by=['data_split'])
#     df = pd.DataFrame({
#         'min_date': gp['date'].min(),
#         'max_date': gp['date'].max(),
#         'size': gp['y_proba'].size(),
#         'count': gp['y_proba'].count(),
#         'pct': float('nan'),
#         'pos_cnt': gp['y'].sum(),
#         'pos_ratio': gp['y'].mean(),
#         'auc': gp.apply(lambda s: roc_auc_score_with_nan(s['y'], s['y_proba'])),
#         'ks': gp.apply(lambda s: ks_score_with_nan(s['y'], s['y_proba']))
#     })
#     df['pct'] = df['size'] / df['size'].sum()
#     data_split_idx = ['train', 'validate', 'test']
#     df = df.loc[data_split_idx, :]
#     df = save_and_reset_index(df)
#     if rename is not None:
#         df['data_split'] = df['data_split'].map(rename)
#     return df


# def get_auc_and_ks_by_month(xy_df: pd.DataFrame):
#     idx = (~xy_df['month'].isna()) & (~xy_df['y'].isna()) & (~xy_df['data_split'].isna())
#     data_df = xy_df.loc[idx, :].copy()
#     gp = data_df.groupby(by=['month'])
#     df = pd.DataFrame({
#         'min_date': gp['date'].min(),
#         'max_date': gp['date'].max(),
#         'size': gp['y_proba'].size(),
#         'count': gp['y_proba'].count(),
#         'coverage': float('nan'),
#         'pct': float('nan'),
#         'pos_cnt': gp['y'].sum(),
#         'pos_ratio': gp['y'].mean(),
#         'auc': gp.apply(lambda s: roc_auc_score_with_nan(s['y'], s['y_proba'])),
#         'ks': gp.apply(lambda s: ks_score_with_nan(s['y'], s['y_proba']))
#     })
#     df['coverage'] = df['count'] / df['size']
#     df['pct'] = df['size'] / df['size'].sum()
#     df = save_and_reset_index(df)
#     return df


# def get_cut_points(s: pd.Series, bins = 20, endpoints = [-float('inf'), float('inf')]):
#     points = []
#     s = pd.Series(s).copy()
#     s.fillna(s.mean(), inplace=True)
#     s.sort_values(inplace=True)
#     for k in range(1, bins):
#         n = int(len(s) * k / bins)
#         cut = s.iloc[n]
#         points.append(cut)
#     n1 = len(points)
#     points = sorted(set(points + endpoints))
#     n2 = len(points)
#     print(f'get_cut_points: info: bins = {n2-1}')
#     if n1 + len(endpoints) != n2:
#         print('get_cut_points: warning: too many duplicated numbers.')
#     points = sorted(points)
#     return points


# def get_gain_and_lift(y, y_proba, bins = 20, prefix = '', y_proba_max = 1):
#     # if bins == 'auto':
#     #     bins = get_cut_points(y_proba, bins=20, endpoints=[0, y_proba_max])
#     bins = get_cut_points(y_proba, bins=bins, endpoints=[0, y_proba_max])
#     bins = sorted(bins, reverse=True)
#     bins = [k for k in bins if k < y_proba_max]
#     print('bins =', bins)
#     df = pd.DataFrame(
#         index=range(1, len(bins)), 
#         columns=[
#             'interval_left', 'interval_right', 
#             'total_count', 'pos_count', 'total_pct', 'pos_ratio', 
#             'gain', 'lift',
#             ],
#         dtype=int)
#     y_mean = y.mean()
#     y_sum = y.sum()
#     for k, cut in enumerate(bins):
#         k = k+1
#         idx = y_proba > cut
#         cut_y = y[idx]
#         cnt = len(cut_y)
#         cnt_rate = cnt / len(y)
#         pos_cnt = cut_y.sum()
#         pos_rate = pos_cnt / cnt
#         lift = pos_cnt / cnt / y_mean
#         # df.loc[k, 'cut_point'] = cut
#         df.loc[k, 'interval_left'] = cut
#         df.loc[k, 'interval_right'] = y_proba_max
#         df.loc[k, 'total_count'] = cnt
#         df.loc[k, 'total_pct'] = cnt_rate
#         df.loc[k, 'gain'] = pos_cnt / y_sum
#         df.loc[k, 'pos_ratio'] = pos_rate
#         df.loc[k, 'pos_count'] = pos_cnt
#         df.loc[k, 'lift'] = lift
#     df = df.loc[df['pos_count'] > 0, :]
#     df = df.add_prefix(prefix)
#     df.insert(0, 'interval', df.apply(lambda row: f"({row['interval_left']:.9f}, {row['interval_right']:.9f}]", axis=1))
#     df.insert(0, 'subset', (((df['total_pct']*100).round()/5).round()*5).apply(lambda s: f'top{s:0.0f}'))
#     df['total_count'] = df['total_count'].astype(int)
#     df['pos_count'] = df['pos_count'].astype(int)
#     df.index.name = 'index'
#     print('lift =')
#     print(df)
#     return df



# def get_gain_and_lift_by_data_split(xy_df: pd.DataFrame, rename = None, y_proba_max = 1, bins = 20):
#     idx = (~xy_df['y_proba'].isna()) & (~xy_df['y'].isna())
#     data_df = xy_df.loc[idx, :]
#     dfs = []
#     for split in ['train', 'validate', 'test']:
#         split_df = data_df.loc[data_df['data_split']==split, :]
#         lift_df = get_gain_and_lift(split_df['y'], split_df['y_proba'], y_proba_max = y_proba_max, bins=bins)
#         lift_df.insert(0, 'data_split', split)
#         dfs.append(lift_df)
#     df = pd.concat(dfs)
#     df.index = pd.Index(range(1, len(df)+1), name='index')
#     if rename is not None:
#         df['data_split'] = df['data_split'].map(rename)
#     return df


# def get_gain_and_lift_by_month(xy_df: pd.DataFrame, y_proba_max = 1, bins = 20):
#     idx = (~xy_df['month'].isna()) & (~xy_df['y'].isna()) & (~xy_df['data_split'].isna()) & (~xy_df['y_proba'].isna())
#     data_df = xy_df.loc[idx, :]
#     dfs = []
#     for month in sorted(data_df['month'].unique()):
#         split_df = data_df.loc[data_df['month']==month, :]
#         lift_df = get_gain_and_lift(split_df['y'], split_df['y_proba'], y_proba_max = y_proba_max, bins=bins)
#         lift_df.insert(0, 'month', month)
#         dfs.append(lift_df)
#     df = pd.concat(dfs)
#     df.index = pd.Index(range(1, len(df)+1), name='index')
#     return df


# def get_binning_table_by_data_split(xy_df, y_proba_max = 1, rename = None, bins=20):
#     valid_idx = (~xy_df['data_split'].isna()) & (~xy_df['y'].isna()) & (~xy_df['y_proba'].isna())
#     xy_df = xy_df.loc[valid_idx, :]
#     dfs = []
#     for split in ['train', 'validate', 'test']:
#         part_df = xy_df.loc[xy_df['data_split']==split, :].copy()
#         bins = get_cut_points(part_df['y_proba'], bins=bins, endpoints=[0, y_proba_max])
#         part_df['interval'] = pd.cut(part_df['y_proba'], bins = bins)
#         gp = part_df.groupby(by='interval', observed=True)
#         part_stat_df = pd.DataFrame({
#             'size': gp.size(),
#             'pct': gp.size() / len(part_df),
#             'min': gp['y_proba'].min(),
#             'max': gp['y_proba'].max(),
#             'pos_ratio': gp['y'].mean()
#         })
#         part_stat_df = save_and_reset_index(part_stat_df, )
#         part_stat_df.sort_index(inplace=True)
#         part_stat_df.insert(0, 'data_split', split)
#         dfs.append(part_stat_df)
#     stat_df = pd.concat(dfs)
#     stat_df.index = pd.Index(range(1, len(stat_df)+1), name = 'index')
#     if rename is not None:
#         stat_df['data_split'] = stat_df['data_split'].map(rename)
#     return stat_df


# def get_binning_table_by_month(xy_df, y_proba_max = 1, bins=20):
#     idx = (~xy_df['month'].isna()) & (~xy_df['y'].isna()) & (~xy_df['data_split'].isna()) & (~xy_df['y_proba'].isna())
#     xy_df = xy_df.loc[idx, :]
#     dfs = []
#     months = sorted(xy_df['month'].unique())
#     for month in months:
#         part_df = xy_df.loc[xy_df['month']==month, :].copy()
#         bins = get_cut_points(part_df['y_proba'], bins=bins, endpoints=[0, y_proba_max])
#         part_df['interval'] = pd.cut(part_df['y_proba'], bins = bins)
#         gp = part_df.groupby(by='interval', observed=True)
#         part_stat_df = pd.DataFrame({
#             'size': gp.size(),
#             'pct': gp.size() / len(part_df),
#             'min': gp['y_proba'].min(),
#             'max': gp['y_proba'].max(),
#             'pos_ratio': gp['y'].mean()
#         })
#         part_stat_df = save_and_reset_index(part_stat_df, )
#         part_stat_df.sort_index(inplace=True)
#         part_stat_df.insert(0, 'month', month)
#         dfs.append(part_stat_df)
#     stat_df = pd.concat(dfs)
#     stat_df.index = pd.Index(range(1, len(stat_df)+1), name = 'index')
#     return stat_df


# def get_distribution_by_data_split(xy_df: pd.DataFrame, rename = None, y_proba_max = 1):
#     idx = (~xy_df['data_split'].isna()) #& (~xy_df['y_proba'].isna())
#     data_df = xy_df.loc[idx, :].copy()

#     bins = get_cut_points(data_df['y_proba'], bins=10, endpoints=[0, y_proba_max])
#     gp = data_df.groupby(by=['data_split'])

#     def func(s):
#         sr = pd.cut(s, bins=bins, precision=9).value_counts().sort_values()
#         sr['Not_a_Number'] = s.isna().sum()
#         return sr
    
#     a_df = pd.DataFrame({
#         'count': gp['y_proba'].apply(func),
#         })
#     a_df['pct'] = a_df['count'] / data_df['data_split'].value_counts()
#     a_df.sort_index(inplace=True)

#     dfs = []
#     for split in ['train', 'validate', 'test']:
#         split_df = a_df.loc[split]
#         split_df = split_df.add_prefix(f'{split}_')
#         dfs.append(split_df)
#     b_df = pd.concat(dfs, axis=1)
#     b_df.index.name = 'interval'
#     b_df = save_and_reset_index(b_df)
#     if rename is not None:
#         new_rename = {}
#         new_rename |= {k+'_count': v+'_count' for k, v in rename.items()}
#         new_rename |= {k+'_pct': v+'_pct' for k, v in rename.items()}
#         b_df.rename(new_rename, axis=1, inplace=True)
#     return b_df


# def get_report_dict(xy_df: pd.DataFrame, rename = None, y_proba_max = 1, bins = 20):
#     if isinstance(y_proba_max, str) and y_proba_max == 'auto':
#         y_proba_max = xy_df['y_proba'].max() + 1
#     report_dict = {}
#     report_dict['describe_by_month'] = get_describe_by_month_df(xy_df=xy_df)
#     report_dict['auc_and_ks_by_data_split'] = get_auc_and_ks_by_data_split(xy_df=xy_df, rename=rename)
#     report_dict['auc_and_ks_by_month'] = get_auc_and_ks_by_month(xy_df=xy_df)
#     report_dict['gain_and_lift_by_data_split'] = get_gain_and_lift_by_data_split(xy_df=xy_df, rename=rename, y_proba_max = y_proba_max, bins=bins)
#     report_dict['gain_and_lift_by_month'] = get_gain_and_lift_by_month(xy_df=xy_df, y_proba_max = y_proba_max, bins = bins)
#     report_dict['binning_table_by_data_split'] = get_binning_table_by_data_split(xy_df=xy_df, y_proba_max = y_proba_max, rename = rename, bins = bins)
#     report_dict['binning_table_by_month'] = get_binning_table_by_month(xy_df=xy_df, y_proba_max = y_proba_max, bins = bins)
#     report_dict['distribution_by_data_split'] = get_distribution_by_data_split(xy_df=xy_df, rename=rename, y_proba_max=y_proba_max)
#     return report_dict

